////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                //
//                            Flood pathfinding with 3dgs native paths                            //
//                                                                                                //
//                             use, modify and distribute as you like                             //
//                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////

// data type for casting datastruct pointers
typedef unsigned long uintptr_t;

// some macros for the edge skills' array indexes
#define PF_LENGTH				0
#define PF_WEIGHT				1

// data struct that allocates all the needed values for the pathfinder
#define pfMaxNodes		pfData1 // This variable is used to allocate the max number of nodes, computed into 'path_router' function
#define pfCost				pfData1 // This variable is used to allocate each node cost into 'route_find' function
typedef struct
{
	var pfData1;
	var pfOpenNext;
	var pfOpenPrev;
	var pfCloseNext;
	var pfClosePrev;
	var pfIncluder;
	var pfStep;
	
} PFDATA;

PFDATA *pfNodeData; // This pointer will content an array to the needed data for pathfinding

// data struct that allocates the found route on the 'route_find' function
typedef struct 
{
	var *pfNodes;
	var pfNodesAmount;
	var pfNodeCurrent;
} PFROUTE;

var route_remove ( var pfRoute )
{
	#ifdef PF_DEBUG_MODE
		if ( !pfRoute )
		{
			error ( "null route on route_remove!" );
			return ( NULL );
		}
	#endif
	
	PFROUTE *pfRouteCurrent = (uintptr_t*)pfRoute;
	
	sys_free ( pfRouteCurrent.pfNodes );
	sys_free ( pfRouteCurrent );
	
	return ( 0 );
}

// algorithm that computes the cost of a node in reference to its includer cost and given edges' length and weight
var pfComputeCost ( var pfNode, var *pfEdgeSkills )
{
	return ( pfNodeData[pfNode].pfCost - ( abs ( pfEdgeSkills[PF_LENGTH] ) * maxv ( pfEdgeSkills[PF_WEIGHT], 1 ) * PF_WEIGHT_SCALE ) );
}

// function that finds a route on a path
var route_find ( ENTITY *pfeHolder, var pfNodeStart, var pfNodeEnd )
{
	#ifdef PF_DEBUG_MODE
		if ( !pfeHolder )
		{
			error ( "null entity on route_find!" );
			return ( NULL );
		}
		VECTOR pfvPosition;
		var pfNodeSkills[6];
		if ( !path_getnode ( pfeHolder, pfNodeStart, pfvPosition, pfNodeSkills ) )
		{
			error ( "unknown start point on route_find!" );
			return ( NULL );
		}
		if ( !path_getnode ( pfeHolder, pfNodeEnd, pfvPosition, pfNodeSkills ) )
		{
			error ( "unknown end point on route_find!" );
			return ( NULL );
		}
	#endif
	
	var pfEdgeCurrent; // edge id number
	var pfEdgeCurrentSkills[3]; // edge skills array
	var pfNodeCurrent, pfNodeTemp, pfNodeCheapest, pfNodeNext, pfNodePrev; // node id numbers
	
	BOOL pfIntoList = OFF; // this variable allocates the state of the current node while searching on the lists
	var pfNewCost = 0; // this variable allocates a new cost for an already computed cost of a node
	
	// start the close list as empty
	var pfCloseFirst = 0;
	var pfCloseLast = 0;
	
	// start the open list and include the starting node into it
	var pfOpenFirst = pfNodeStart;
	var pfOpenLast = pfNodeStart;
	pfNodeData[pfOpenFirst].pfCost = 1000000; // set a high value (lowest cost) for the starting node
	pfNodeData[pfOpenFirst].pfOpenNext = 0; // since it is the unique node into the open list it has no neighbors
	pfNodeData[pfOpenFirst].pfOpenPrev = 0;
	pfNodeData[pfOpenFirst].pfIncluder = 0; // starting node has no includer
	pfNodeData[pfOpenFirst].pfStep = 0;
	
	pfNodeData[pfNodeEnd].pfCost = -1000000; // set a low value (highest cost) for the ending node
	
	while ( pfOpenFirst ) // while the open list has an item
	{
		// Look for the cheapest node into the open list
		pfNodeCheapest = pfOpenFirst;
		pfNodeCurrent = pfNodeData[pfNodeCheapest].pfOpenNext;
		while ( pfNodeCurrent )
		{
			if ( pfNodeData[pfNodeCurrent].pfCost > pfNodeData[pfNodeCheapest].pfCost ) // the higher pfCost value the cheaper cost
			{
				pfNodeCheapest = pfNodeCurrent;
			}
			pfNodeCurrent = pfNodeData[pfNodeCurrent].pfOpenNext;
		}
		
		#ifdef SECURE_MODE
			if ( pfNodeData[pfNodeCheapest].pfCost < -990000 )
			{
				error ( "Cost maths dangerously approaches the var range limits.\n Modify PF_WEIGHT_SCALE into '3dgsPathFinder.h' to let it stay into the range!" );
				return ( 0 );
			}
		#endif
		
		// exclude the cheapest node from the open list
		if ( pfNodeData[pfNodeCheapest].pfOpenNext && pfNodeData[pfNodeCheapest].pfOpenPrev ) // does it have both neighbors on the list?
		{
			pfNodeNext = pfNodeData[pfNodeCheapest].pfOpenNext;
			pfNodePrev = pfNodeData[pfNodeCheapest].pfOpenPrev;
			pfNodeData[pfNodePrev].pfOpenNext = pfNodeNext;
			pfNodeData[pfNodeNext].pfOpenPrev = pfNodePrev;
		}
		else if ( pfNodeData[pfNodeCheapest].pfOpenPrev ) // is it the last item on the list?
		{
			pfNodePrev = pfNodeData[pfNodeCheapest].pfOpenPrev;
			pfNodeData[pfNodePrev].pfOpenNext = 0;
			pfOpenLast = pfNodePrev;
		}
		else if ( pfNodeData[pfNodeCheapest].pfOpenNext ) // is it the first item on the list?
		{
			pfNodeNext = pfNodeData[pfNodeCheapest].pfOpenNext;
			pfNodeData[pfNodeNext].pfOpenPrev = 0;
			pfOpenFirst = pfNodeNext;
		}
		else // so, it is the only item on the list
		{
			pfOpenFirst = pfOpenLast = 0;
		}
		
		// include the cheapest node on the close list
		if ( !pfCloseFirst ) // is the close list empty?
		{
			pfNodeData[pfNodeCheapest].pfCloseNext = 0;
			pfNodeData[pfNodeCheapest].pfClosePrev = 0;
			pfCloseFirst = pfCloseLast = pfNodeCheapest;
		}
		else // if not, iclude the cheapest node as first item on the close list
		{
			pfNodeData[pfNodeCheapest].pfClosePrev = 0;
			pfNodeData[pfNodeCheapest].pfCloseNext = pfCloseFirst;
			pfNodeData[pfCloseFirst].pfClosePrev = pfNodeCheapest;
			pfCloseFirst = pfNodeCheapest;
		}
		
		// Include the cheapest nodes' neighborhood on the lists
		pfEdgeCurrent = 1;
		while ( path_getedge ( pfeHolder, pfNodeCheapest, pfEdgeCurrent, pfEdgeCurrentSkills ) ) // Look into all the path edges connected to the cheapest node
		{
			// Retrieve the current edges' other end node
			pfNodeCurrent = path_nextnode ( pfeHolder, pfNodeCheapest, pfEdgeCurrent );
			pfIntoList = OFF;
			
			#ifdef PF_DEBUG_MODE
				if ( pfNodeCurrent > pfNodeData[0].pfMaxNodes - 1 )
				{
					error ( "The current node id number is out of pathfinder data range. \n Call 'path_router(ENTITY* ent)' for each entity that uses the pathfinder." );
					return ( 0 );
				}
			#endif
			
			if ( pfNodeData[pfNodeCheapest].pfIncluder != pfNodeCurrent ) // In the case that the other end node is not the includer of the cheapest node
			{
				// look for the other end node on the close list
				pfNodeTemp = pfCloseFirst;
				
				while ( pfNodeTemp )
				{
					if ( pfNodeTemp == pfNodeCurrent ) // if the other end node is on the close list
					{
						pfIntoList = ON; // so, it is on a list
						
						pfNewCost = pfComputeCost ( pfNodeCheapest, pfEdgeCurrentSkills ); // compute the current cost
						
						if ( pfNewCost > pfNodeData[pfNodeCurrent].pfCost ) // if the new cost is cheaper than the old one...
						{
							// set the new cost for the node
							pfNodeData[pfNodeCurrent].pfCost = pfNewCost; 
							
							// exclude the node from the close list
							if ( pfNodeData[pfNodeCurrent].pfClosePrev && pfNodeData[pfNodeCurrent].pfCloseNext ) // does it have both neighbors on the list?
							{
								pfNodeNext = pfNodeData[pfNodeCurrent].pfCloseNext;
								pfNodePrev = pfNodeData[pfNodeCurrent].pfClosePrev;
								pfNodeData[pfNodeNext].pfClosePrev = pfNodePrev;
								pfNodeData[pfNodePrev].pfCloseNext = pfNodeNext;
							}
							else if ( pfNodeData[pfNodeCurrent].pfClosePrev ) // is it the last item on the list?
							{
								pfNodePrev = pfNodeData[pfNodeCurrent].pfClosePrev;
								pfNodeData[pfNodePrev].pfCloseNext = 0;
								pfCloseLast = pfNodePrev;
							}
							else if ( pfNodeData[pfNodeCurrent].pfCloseNext ) // is it the first item on the list?
							{
								pfNodeNext = pfNodeData[pfNodeCurrent].pfCloseNext;
								pfNodeData[pfNodeNext].pfClosePrev = 0;
								pfCloseFirst = pfNodeNext;
							}
							else // so, it is the only item on the list
							{
								pfCloseFirst = pfCloseLast = 0;
							}
							
							// and include it into the open list
							if ( !pfOpenLast ) // is the open list empty?
							{
								pfNodeData[pfNodeCurrent].pfOpenNext = 0;
								pfNodeData[pfNodeCurrent].pfOpenPrev = 0;
								pfNodeData[pfNodeCurrent].pfIncluder = pfNodeCheapest;
								pfNodeData[pfNodeCurrent].pfStep = pfNodeData[pfNodeCheapest].pfStep + 1;
								pfOpenLast = pfOpenFirst = pfNodeCurrent;
							}
							else // if not, iclude the other end node as last item on the open list
							{
								pfNodeData[pfOpenLast].pfOpenNext = pfNodeCurrent;
								pfNodeData[pfNodeCurrent].pfOpenPrev = pfOpenLast;
								pfNodeData[pfNodeCurrent].pfOpenNext = 0;
								pfNodeData[pfNodeCurrent].pfIncluder = pfNodeCheapest;
								pfNodeData[pfNodeCurrent].pfStep = pfNodeData[pfNodeCheapest].pfStep + 1;
								pfOpenLast = pfNodeCurrent;
							}
						}
						
						break; // the other end node is on the close list, so stop searching
					}
					
					pfNodeTemp = pfNodeData[pfNodeTemp].pfCloseNext;
				}
				
				// if the edges' other end node is not on the close list, look for it on the open list
				if ( !pfIntoList )
				{
					pfNodeTemp = pfOpenFirst;
					
					while ( pfNodeTemp )
					{
						if ( pfNodeTemp == pfNodeCurrent ) // if the edges' other end node is on the open list
						{
							pfIntoList = ON; // so, it is on a list
							
							pfNewCost = pfComputeCost ( pfNodeCheapest, pfEdgeCurrentSkills );  // compute the current cost
							
							if ( pfNewCost > pfNodeData[pfNodeCurrent].pfCost ) // if the new cost is cheaper than the old one...
							{
								// set the new cost for the node
								pfNodeData[pfNodeCurrent].pfCost = pfNewCost;
								pfNodeData[pfNodeCurrent].pfIncluder = pfNodeCheapest;
								pfNodeData[pfNodeCurrent].pfStep = pfNodeData[pfNodeCheapest].pfStep + 1;
							}
							
							break; // the edges' other end node is on the open list, so stop searching
						}
						
						pfNodeTemp = pfNodeData[pfNodeTemp].pfOpenNext; // go to next item on the open list
					}
				}
				
				// if the edges' other end node is not on a list, include it on the open list
				if ( !pfIntoList )
				{
					if ( !pfOpenLast ) // is the open list empty?
					{
						pfNodeData[pfNodeCurrent].pfCost = pfComputeCost ( pfNodeCheapest, pfEdgeCurrentSkills );
						pfNodeData[pfNodeCurrent].pfOpenNext = 0;
						pfNodeData[pfNodeCurrent].pfOpenPrev = 0;
						pfNodeData[pfNodeCurrent].pfIncluder = pfNodeCheapest;
						pfNodeData[pfNodeCurrent].pfStep = pfNodeData[pfNodeCheapest].pfStep + 1;
						pfOpenFirst = pfOpenLast = pfNodeCurrent;
					}
					else // if not, iclude the edges' other end node as the last item on the open list
					{
						pfNodeData[pfOpenLast].pfOpenNext = pfNodeCurrent;
						pfNodeData[pfNodeCurrent].pfCost = pfComputeCost ( pfNodeCheapest, pfEdgeCurrentSkills );
						pfNodeData[pfNodeCurrent].pfOpenPrev = pfOpenLast;
						pfNodeData[pfNodeCurrent].pfOpenNext = 0;
						pfNodeData[pfNodeCurrent].pfIncluder = pfNodeCheapest;
						pfNodeData[pfNodeCurrent].pfStep = pfNodeData[pfNodeCheapest].pfStep + 1;
						pfOpenLast = pfNodeCurrent;
					}
				}
			}
			
			pfEdgeCurrent ++; // go to next edge
		}
		
		// the cheapest route is found when:
		// the cheapest nodes' cost is higher (its value is smaller) than the end node cost
		// or when the cheapest node is the end node and the open list is empty
		if ( ( pfNodeData[pfNodeCheapest].pfCost < pfNodeData[pfNodeEnd].pfCost ) || ( ( pfNodeCheapest == pfNodeEnd ) && !pfOpenFirst ) )
		{
			PFROUTE *pfRoute = sys_malloc(sizeof(PFROUTE)); // allocate the needed memory
			pfRoute.pfNodesAmount = pfNodeData[pfNodeEnd].pfStep + 1; 
			pfRoute.pfNodeCurrent = 0; 
			pfRoute.pfNodes = sys_malloc(pfRoute.pfNodesAmount*sizeof(var));
			
			pfNodeCurrent = pfNodeEnd;
			while ( pfNodeCurrent != pfNodeStart )
			{
				(pfRoute.pfNodes)[pfNodeData[pfNodeCurrent].pfStep] = pfNodeCurrent;
				pfNodeCurrent = pfNodeData[pfNodeCurrent].pfIncluder;
			}
			
			(pfRoute.pfNodes)[0] = pfNodeStart;
			
			return ( (var)pfRoute );
		}
	}
	
	// since the open list is empty without finding a route...
	return ( 0 ); // ...there is no way
}

// this function computes the node amount of the path attached to the entity and returns its value
// also resizes the array used for pathfinding, in the case that it needs be bigger than the already allocated array
var path_router ( ENTITY *pfeHolder )
{
	// variables needed for retrieving path nodes
	var pfNodes = 1; 	// since first path node is indexed as 1 and the pathfinder uses its index number into a data struct array, 
							// the pathfinder will need an array one unit bigger than the nodes amount
	var pfNodeSkills[6];
	VECTOR pfvPosition;
	
	#ifdef PF_DEBUG_MODE
		if ( !pfeHolder )
		{
			error ( "null entity on path_router!" );
			return ( 0 );
		}
		if ( !path_getnode ( pfeHolder, pfNodes, pfvPosition, pfNodeSkills ) )
		{
			error ( "entity does not have an attached path on path_router!" );
			return ( 0 );
		}
	#endif
	
	
	// go through path nodes to reach the last
	while ( path_getnode ( pfeHolder, pfNodes, pfvPosition, pfNodeSkills ) )
	{
		pfNodes ++; // it always is one unit bigger than the real nodes amount
	}
	
	// pfNodeData[0].pfMaxNodes allocates the pfNodeData array size
	if ( pfNodeData[0].pfMaxNodes < pfNodes )
	{
		sys_free ( pfNodeData );
		
		pfNodeData = sys_malloc(pfNodes*sizeof(PFDATA));
		pfNodeData[0].pfMaxNodes = pfNodes;
	}
	
	return ( pfNodes - 1 ); // The real nodes amount is one less than the computed amount
}

// This function manages a found route.
var route_nextnode ( var pfRoute )
{
	#ifdef PF_DEBUG_MODE
		if ( !pfRoute )
		{
			error ( "null route on route_nextnode!" );
			return ( NULL );
		}
	#endif
	
	PFROUTE *pfRouteCurrent = (uintptr_t*)pfRoute;
	
	pfRouteCurrent.pfNodeCurrent ++;
	
	if ( pfRouteCurrent.pfNodeCurrent > pfRouteCurrent.pfNodesAmount )
		return ( 0 );
	else
		return ( (pfRouteCurrent.pfNodes)[pfRouteCurrent.pfNodeCurrent] );
}

// This function restarts the route, and returns the first node id number.
var route_restart ( var pfRoute )
{
	#ifdef PF_DEBUG_MODE
		if ( !pfRoute )
		{
			error ( "null route on route_restart!" );
			return ( NULL );
		}
	#endif
	
	PFROUTE *pfRouteCurrent = (uintptr_t*)pfRoute;
	
	pfRouteCurrent.pfNodeCurrent = 0;
	
	return ( (pfRouteCurrent.pfNodes)[0] );
}

// This function clones a route for using it by different entities or functions.
var route_clone ( var pfRoute )
{
	#ifdef PF_DEBUG_MODE
		if ( !pfRoute )
		{
			error ( "null route on route_clone!" );
			return ( NULL );
		}
	#endif
	
	PFROUTE *pfRouteCurrent = (uintptr_t*)pfRoute;
	
	PFROUTE *pfRouteNew = sys_malloc(sizeof(PFROUTE));
	pfRouteNew.pfNodesAmount = pfRouteCurrent.pfNodesAmount;
	pfRouteNew.pfNodeCurrent = 0; 
	pfRouteNew.pfNodes = sys_malloc(pfRouteCurrent.pfNodesAmount*sizeof(var));
	memcpy ( &(pfRouteNew.pfNodes)[0], &(pfRouteCurrent.pfNodes)[0], pfRouteCurrent.pfNodesAmount*sizeof(var) );
	
	return ( (var)pfRouteNew );
}

// This function allocates a minimum size for the array used for the pathfinding, executed at game start
function pf_startup ()
{
	pfNodeData = sys_malloc(sizeof(PFROUTE)); // allocate the memory
	pfNodeData[0].pfMaxNodes = 0; // save the nodes amount
}

// This function deallocates the needed memory for the pathfinder
function pf_close ()
{
	sys_free ( pfNodeData );
}

