////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                //
//                            Flood pathfinding with 3dgs native paths                            //
//                                                                                                //
//                             use, modify and distribute as you like                             //
//                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////

// HOW THIS PATHFINDER WORKS
// -------------------------------------------------------------------------------------------------

// This pathfinder uses two linked lists for finding the cheapest route from one node to an other node on a 3dgs PATH struct.
// Some of you may know them: They are the open and the close lists.

// The open list contains all the nodes that are sensitive to be part of the searched route.
// The close list contains all the nodes that are already computed.

// The algorithm searches for the cheapest node on the open list, moves it to the close list 
// and includes its neighborhood on the open list. Every neighbor of the cheapest node is searched
// on both list before including on the open list and actualized if it is necessary.

// The cost of a node (the value that lets the algorithm know the cheapest one) is computed when
// the node is included on the open list. It also saves the node that includes the opened
// node and its step number, so it facilitates the route build when the end node is reached.

// It starts with an empty close list and includes the starting node on the open list.

// The pathfinder loop edits both lists until reaching the end node, but it does not finish there...
// It continues until the cheapest node is expensiver than the already computed end node, because there are some cases where
// the end node can be earlier reached by a expensiver path than the cheapest one.

// If the end node is never reached, because there is no connection between starting and ending node,
// the open list gets empty and the pathfinder returns a zero value.

// If the end node is properly reached, the route nodes are known because of their includer starting from the end node.
// The pathfinder function returns a casted pointer to a struct that contains an array with the nodes
// ordered from starting node to ending node. 
// You can manage this struct with some functions explained into the next chapter.

// You will find more explicit explanations into '3dgsPathFinder.c' file, where all happens.



// CONFIGURATION
// -------------------------------------------------------------------------------------------------

// #define PF_DEBUG_MODE
// 
// This define enables some pointers checks, but it is not so secure. It will fail on its job if you give as parameter an invalid pointer...
// Enable it for recognising a detectable error into pathfinder functions


#define PF_WEIGHT_SCALE 		1 		// PF_WEIGHT_SCALE > 0
// 
// This macro is used to modify how the cost of an edge is calculated.
// It is just a factor that multiplies the edges' lenght and weight,
// so it has to be bigger than zero.
// Since the var type range goes nearly from -1000000 to 1000000, it may be used to scale the edges' lenght when very big paths or very big edge weights are used



// HOW TO USE THIS PATHFINDER
// -------------------------------------------------------------------------------------------------

// First of all you will need a PATH struct drawn in WED (or created dynamically) and attached to an entity.

// All the paths (at least the biggest one) that you want to use for pathfinding needs to be computed with:

var path_router ( ENTITY *pfeHolder ); // where pfeHolder is an entity with an attached path.

// This function computes and returns the nodes amount of the attached path, and ensures that the data struct used for the pathfinding is big enough. 
// You can save this value in a var type variable if you want to.
// fe: var NodesAmount = path_router ( myEntity );
// or: my.skill1 = path_router ( me );
// Don't forget calling this function for each path after charging a new level or after modifying a path dynamically,
// otherwise you can get some memory overwriting problems.


// It's time to find a route

var route_find ( ENTITY *pfeHolder, var pfNodeStart, var pfNodeEnd );

// It is somehow self explanatory.
// pfeHolder is the entity that contains the path.
// pfNodeStart is the starting node for the route.
// pfNodeEnd is the end node

// You should be sure that both nodes are existing nodes into the path.
// Use 3dgs funtions, as path_scanpath, for this purpouse.

// It returns a unique route id. More exactly a casted pointer to a data struct that allocates all the nodes numbers.
// If there is no way between given nodes, it returns 0.
// You must save this value into a var or entity skill for further manipulation.
// fe: var myRoute = route_find ( me, 5, 22 );
// or: my.skill2 = route_find ( myPathHolder, 14, 3 );


// EVERY ROUTE HAS TO BE REMOVED WHEN IT IS NO LONGER NEEDED
// Use this function for this purpouse:

var route_remove ( var pfRoute );

// This function always returns zero, so you can use it for reseting your route holder var.
// fe: my.skill1 = route_remove ( my.skill1 );

// You should be sure that the route exist, otherwise you will get a big crash.


// This function manages a found route.

var route_nextnode ( var pfRoute );

// It returns the next node into the route or zero if there are no more nodes,
// so use this var for retrieving the node values with path_getnode.
// You know the first node that is used into route_find function,
// so the first node it returns is the next node to the first.


// This function restarts the route, and returns the first node id number.

var route_restart ( var pfRoute );


// This function clones a route for using it by different entities or functions.

var route_clone ( var pfRoute );

// It returns a new unique id for the copy.
// The pointed struct returned by this function should be removed too!


// This function allocates a minimum size for the array used for the pathfinding

function pf_startup ();

// It is executed at the program start.


// This function deallocates the needed memory for the pathfinder

function pf_close ();

// Call it at the end of your program


// include the interesting code
#include "3dgsPathFinder.c"


// Have fun ;)
