////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                //
//                            Flood pathfinding with 3dgs native paths                            //
//                                                                                                //
//                             use, modify and distribute as you like                             //
//                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////

// You will find some instructions into '3dgsPathFinder.h' file
// and some examples into this file

#include <acknex.h>
#include <default.c>

#include "3dgsPathFinder.h"

VECTOR vtemp;
ANGLE atemp;
var ntemp;

BMAP *IMG_Mouse = "a.bmp"; // mouse pointer

// this function draws the edges and nodes of the path attached to an entity
function draw_path ( ENTITY *ENT_Holder )
{ 
	// for retrieving the nodes and edges data
	VECTOR VEC_Position, VEC_Position2;
	var NodeCurrent=1, NodeSkills[6];
	var EdgeCurrent, EdgeSkills[3];
	
	while ( path_getnode ( ENT_Holder, NodeCurrent, VEC_Position, NodeSkills ) ) // go through all nodes into the path
	{
		EdgeCurrent = 1;
		
		draw_point3d ( VEC_Position, COLOR_BLUE, 100, 10 );
		draw_line3d ( VEC_Position, COLOR_BLUE, 0 );
		
		while ( path_getedge ( ENT_Holder, NodeCurrent, EdgeCurrent, EdgeSkills ) ) // go through all edges connected to the current node
		{
			path_getnode ( ENT_Holder, path_nextnode ( ENT_Holder, NodeCurrent, EdgeCurrent ), VEC_Position2, NodeSkills ); // retrieve the other end node position
			
			vec_sub ( VEC_Position2, VEC_Position ); // show the unidirectional edges by cutting its length
			vec_scale ( VEC_Position2, 0.75 );
			vec_add ( VEC_Position2, VEC_Position );
			
			draw_line3d ( VEC_Position, COLOR_BLUE, 100 );
			draw_line3d ( VEC_Position2, COLOR_BLUE, 100 );
			
			EdgeCurrent ++;
		}
		
		draw_line3d ( VEC_Position2, COLOR_BLUE, 0 );
		
		NodeCurrent ++;
	}
}

// this function draws the handlers
function draw_handler ( VECTOR *vHandler, var PathNode, COLOR *Color )
{
	draw_point3d ( vHandler, Color, 100, 30 );
	draw_line3d ( vHandler, Color, 0 );
	draw_line3d ( vHandler, Color, 100 );
	VECTOR vPosition;
	path_getnode ( me, PathNode, vPosition, NULL );
	draw_line3d ( vPosition, Color, 100 );
	draw_line3d ( vPosition, Color, 0 );
}

// This function draws a route
function draw_route ( ENTITY *PathHolder, var Route, COLOR *Color )
{
	VECTOR VEC_Position;
	var NodeSkills[6];
	
	var Node = route_restart ( Route ); // restart route and retrieve first node of the route
	
	path_getnode ( PathHolder, Node, VEC_Position, NodeSkills );
	draw_line3d ( VEC_Position, Color, 0 );
	draw_line3d ( VEC_Position, Color, 100 );
	
	Node = route_nextnode ( Route ); // retrieve next node
	
	while ( Node )
	{
		path_getnode ( PathHolder, Node, VEC_Position, NodeSkills );
		draw_line3d ( VEC_Position, Color, 100 );
		Node = route_nextnode ( Route ); // retrieve next node of the route
	}
	
	draw_line3d ( VEC_Position, Color, 0 );
}

// This function spawns moving points along a route
function route_spawnpoint ( ENTITY *PathHolder, var Route, COLOR *Color )
{
	var RouteLocal = route_clone ( Route ); // make a local copy of the route for using it without danger of deletion
	
	var NodeCurrent = route_restart ( RouteLocal ); // restart route and retrieve first node of the route
	var NodeNext = route_nextnode ( RouteLocal ); // retrieve next node
	
	var NodeSkills[6];
	
	VECTOR vStartPoint;
	VECTOR vEndPoint;
	VECTOR vSegment;
	
	COLOR ColorLocal;
	vec_set ( ColorLocal, Color );
	
	var Time = 0;
	var Clock = total_ticks;
	
	while ( NodeNext ) // route_nextnode returns zero when finished
	{
		path_getnode ( PathHolder, NodeCurrent, vStartPoint, NodeSkills );
		path_getnode ( PathHolder, NodeNext, vEndPoint, NodeSkills );
		vec_diff ( vSegment, vEndPoint, vStartPoint );
		
		Clock += Time;
		Time = vec_length ( vSegment ) / 20;
		
		while ( total_ticks - Clock < Time )
		{
			vec_set ( vtemp, vSegment );
			vec_scale ( vtemp, ( total_ticks - Clock ) / Time );
			vec_add ( vtemp, vStartPoint );
			draw_point3d ( vtemp, ColorLocal, 100, 15 );
			
			wait(1);
		}
		
		NodeCurrent = NodeNext;
		NodeNext = route_nextnode ( RouteLocal ); // retrieve next node of the route
	}
	
	route_remove ( RouteLocal ); // remove the local route
}


// Renaming skills with defines
#define AmountOfNodes		skill99		
#define Route					skill98		

// action attached to the path holder entity
action GSPATH_Holder ()
{
	set ( me, PASSABLE | INVISIBLE );
	
	VECTOR VEC_StartPoint, VEC_EndPoint, VEC_Position;
	vec_set ( VEC_StartPoint, vector(-360,306,0) );
	vec_set ( VEC_EndPoint, vector(570,56,0) );
	var StartPoint = path_scanpath ( me, VEC_StartPoint, my.pan, vector(360,180,1000) );
	var EndPoint = path_scanpath ( me, VEC_EndPoint, my.pan, vector(360,180,1000) );
	
// THIS IS A MUST! ---------------------- //
	my.AmountOfNodes = path_router ( me ); //
// -------------------------------------- //
	
	my.Route = route_find ( me, StartPoint, EndPoint ); // Find a route
	
	var Clock = total_ticks;
	
	while ( !key_esc )
	{
		if ( mouse_left )
		{
			// find the projection of the mouse over the z=0 horizontal plane
			vec_set ( vtemp, mouse_dir3d );
			if ( vtemp.z ) vec_scale(vtemp,-camera.z/vtemp.z);
			vec_add ( vtemp, camera.x );
			vec_set ( VEC_StartPoint, vtemp );
			
			// find the nearest node
			StartPoint = path_scanpath ( me, VEC_StartPoint, my.pan, vector(360,180,1000) );
			
			// recreate the route
			if ( my.Route ) my.Route = route_remove ( my.Route ); // Remove a previous route before finding a new one
			if ( StartPoint && EndPoint ) my.Route = route_find ( me, StartPoint, EndPoint ); // Find a new route
		}
		
		if ( mouse_right )
		{
			// find the projection of the mouse over the z=0 horizontal plane
			vec_set ( vtemp, mouse_dir3d );
			if ( vtemp.z ) vec_scale ( vtemp, -camera.z/vtemp.z );
			vec_add ( vtemp, camera.x );
			vec_set ( VEC_EndPoint, vtemp );
			
			// find the nearest node
			EndPoint = path_scanpath ( me, VEC_EndPoint, my.pan, vector(360,180,1000) );
			
			// recreate the route
			if ( my.Route ) my.Route = route_remove ( my.Route ); // Remove a previous route before finding a new one
			if ( StartPoint && EndPoint ) my.Route = route_find ( me, StartPoint, EndPoint ); // Find a new route
		}
		
		if ( my.Route )
		{
			// draw the route
			draw_route ( me, my.Route, COLOR_WHITE );
			
			// spawn a point along the route every eight ticks
			if ( total_ticks - Clock > 8 )
			{
				ntemp = integer ( random ( 255 ) );
				route_spawnpoint ( me, my.Route, vector ( 0, ntemp, 255 ) ); 
				Clock = total_ticks;
			}
		} 
		
		// draw other stuff
		draw_handler ( VEC_StartPoint, StartPoint, COLOR_WHITE );
		draw_handler ( VEC_EndPoint, EndPoint, COLOR_GREEN );
		draw_path ( me );
		
		draw_text ( "draw paths in WED and find the way", 240, 280, COLOR_WHITE );
		draw_text ( "mouse_left = start point", 732, 600, COLOR_WHITE );
		draw_text ( "mouse_right = end point", 730, 620, COLOR_GREEN );
		draw_text ( "direction and weight", 740, 155, COLOR_RED );
		
		wait(1);
	}
	
	if ( my.Route ) my.Route = route_remove ( my.Route );
}


function main()
{
	video_mode = 8;
	video_window ( 0, 0, 0, "Flood pathfinding with 3dgs native paths" );
	
	on_esc = NULL;
	
	level_load ( "l.wmb" );
	
	ent_createlayer ( "s.tga", SKY | CUBE, 1 );
	
	mouse_map = IMG_Mouse;
	mouse_mode = 4;
	
	while ( !key_esc ) wait(1);
	
	wait(1);
	
	sys_exit ( NULL );
}



